from typing import Any, Hashable, List, Sequence, Tuple, Type, TypeVar

import cooperative_dataclasses as dataclasses
from jax.tree_util import register_pytree_node

from .annotations import PyTree
from .display import display_class, display_key_and_value
from .pytree_like import PyTreeLike

__all__ = ['dataclass', 'field']


T = TypeVar('T', bound=Any)


def dataclass(clz: Type[T]) -> Type[T]:
    """
    some_member: SomeType = field(pytree_like=False)
    """
    # pylint: disable=protected-access

    # Apply dataclass function to clz.
    data_clz: Type[T] = dataclasses.dataclass(frozen=True)(clz)  # type: ignore

    # Partition fields into hashed, tree, and uninitialized.
    hashed_fields: List[str] = []
    tree_fields: List[str] = []
    for field_info in dataclasses.fields(data_clz):  # type: ignore
        if not field_info.init:
            continue
        if field_info.metadata.get('pytree_like', True):
            tree_fields.append(field_info.name)
        else:
            hashed_fields.append(field_info.name)

    # Generate additional methods.
    def __repr__(self: T) -> str:
        return str(self.display())

    def display(self: T, show_values: bool = True, indent: int = 0) -> str:
        retval = display_class(type(self))
        for field_info in dataclasses.fields(data_clz):  # type: ignore
            retval += display_key_and_value(
                field_info.name, getattr(self, field_info.name), "=", show_values, indent)
        return retval

    def tree_flatten(x: T) -> Tuple[Sequence[PyTree], Hashable]:
        hashed = tuple(getattr(x, name) for name in hashed_fields)
        trees = tuple(getattr(x, name) for name in tree_fields)
        return trees, hashed

    def tree_unflatten(cls: Type[T], hashed: Hashable, trees: Sequence[PyTree]) -> T:
        if not isinstance(hashed, tuple):
            raise TypeError
        hashed_args = dict(zip(hashed_fields, hashed))
        tree_args = dict(zip(tree_fields, trees))
        return cls(**hashed_args, **tree_args)

    # Assign methods to the class.
    data_clz.__repr__ = __repr__  # type: ignore
    data_clz.display = display  # type: ignore
    data_clz.tree_flatten = tree_flatten  # type: ignore
    data_clz.tree_unflatten = classmethod(tree_unflatten)  # type: ignore

    # Assign field lists to the class.
    data_clz.tree_fields = tree_fields  # type: ignore
    data_clz.hashed_fields = hashed_fields  # type: ignore

    # Register the class as a JAX PyTree.
    register_pytree_node(data_clz, tree_flatten, data_clz.tree_unflatten)  # type: ignore

    # Verify that the generated class is PyTreeLike.
    assert isinstance(data_clz, PyTreeLike)

    return data_clz


def field(pytree_like: bool = True, **kwargs: Any) -> dataclasses.Field:
    return dataclasses.field(metadata={**kwargs.pop('metadata', {}),
                                       'pytree_like': pytree_like},
                             **kwargs)



"""
This package provides abstract interfaces:
* tensor manipulation (TensorManipulator),
* interval arithmetic (Extrema), and
* slicing and broadcasting (Matching).

The interface is implemented using cpu and gpu routines.
"""
from .annotations import *
from .dataclass import *
from .display import *
from .dtypes import *
from .generator import *
from .graph import *
from .pytree_like import *
from .shims import *
from .testing import *
from .tools import *



from typing import Any, Callable, Generic, Tuple, TypeVar, Union

from jax import custom_vjp as jax_custom_vjp
from jax import numpy as jnp
from jax.tree_util import tree_map

__all__ = ['custom_vjp']


R = TypeVar('R')


def as_tuple(x: Union[int, Tuple[int, ...]]) -> Tuple[int, ...]:
    if isinstance(x, int):
        return (x,)
    return x


class custom_vjp(Generic[R]):
    """
    Shim class to work around https://github.com/google/jax/issues/2912.
    """
    def __init__(self,
                 fun: Callable[..., R],
                 static_argnums: Union[int, Tuple[int, ...]] = (),
                 nondiff_argnums: Union[int, Tuple[int, ...]] = ()):
        static_argnums = as_tuple(static_argnums)
        nondiff_argnums = as_tuple(nondiff_argnums)
        if intersection := set(static_argnums) & set(nondiff_argnums):
            raise ValueError(
                f"Arguments {intersection} cannot be both static and nondifferentiable.")
        self.nondiff_argnums = nondiff_argnums
        self.vjp = jax_custom_vjp(fun, nondiff_argnums=static_argnums)

    def defvjp(self, fwd: Callable[..., Tuple[R, Any]], bwd: Callable[..., Any]) -> None:
        def new_fwd(*args: Any) -> Tuple[R, Any]:
            zeroed_args = tuple([tree_map(jnp.zeros_like, args[i])
                                 for i in self.nondiff_argnums])
            primal, internal_residuals = fwd(*args)
            return primal, (zeroed_args, internal_residuals)

        def new_bwd(residuals: Any, output_bar: R) -> Any:
            zeroed_args, internal_residuals = residuals
            input_bar = bwd(internal_residuals, output_bar)
            input_bar = list(input_bar)
            for i, index in enumerate(self.nondiff_argnums):
                input_bar[index: index] = [zeroed_args[i]]
            return tuple(input_bar)

        self.vjp.defvjp(new_fwd, new_bwd)

    def __call__(self, *args: Any) -> R:
        return self.vjp(*args)

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R]:
        # https://github.com/google/jax/issues/2483
        return self.vjp.__get__(instance, owner)



from functools import partial
from numbers import Number
from typing import Any, Optional, cast

import numpy as np
from jax import numpy as jnp
from jax.interpreters.xla import DeviceArray
from jax.tree_util import tree_multimap, tree_reduce

from .annotations import PyTree
from .dtypes import default_atol, default_rtol

__all__ = ['assert_jax_allclose', 'jax_allclose', 'get_test_string']


def assert_jax_allclose(actual: PyTree,
                        desired: PyTree,
                        original_name: Optional[str] = None,
                        original_value: Optional[PyTree] = None,
                        *,
                        rtol: Optional[float] = None,
                        atol: Optional[float] = None) -> None:
    if rtol is None:
        rtol = default_rtol
    if atol is None:
        atol = default_atol

    try:
        tree_multimap(partial(np.testing.assert_allclose, rtol=rtol, atol=atol), actual, desired)
    except:
        print("JAX trees don't match.  Actual:")
        print(actual)
        print("Desired:")
        print(desired)
        if original_name is not None and original_value is not None:
            print("Test string:")
            print(get_test_string(original_name, actual, original_value, rtol, atol))
        raise


def jax_allclose(actual: PyTree,
                 desired: PyTree,
                 rtol: Optional[float] = None,
                 atol: Optional[float] = None) -> bool:
    if rtol is None:
        rtol = default_rtol
    if atol is None:
        atol = default_atol

    return cast(
        bool,
        tree_reduce(jnp.logical_and,
                    tree_multimap(partial(np.allclose, rtol=rtol, atol=atol), actual, desired),
                    True))


def get_test_string(original_name: str,
                    actual: Any,
                    original: Any,
                    rtol: Optional[float] = None,
                    atol: Optional[float] = None) -> str:
    if isinstance(actual, (np.ndarray, DeviceArray)):
        return "np." + repr(np.asarray(actual))
    if isinstance(actual, Number):
        return str(actual)
    if hasattr(actual, 'display'):
        retval = f"{original_name}.replace("
        retval += ",\n".join(
            f"{fn}=" + get_test_string(f"{original_name}.{fn}",
                                       getattr(actual, fn),
                                       getattr(original, fn),
                                       rtol,
                                       atol)
            for fn in actual.tree_fields
            if not jax_allclose(getattr(actual, fn), getattr(original, fn), rtol=rtol, atol=atol))
        retval += ")"
        return retval
    if isinstance(actual, dict):
        return str({key: get_test_string(f"{original_name}[{key}]", sub_value, original[key],
                                         rtol, atol)
                    for key, sub_value in actual.items()})
    return str(actual)



import importlib
from typing import Any, List

import colorful as cf
from arpeggio import OneOrMore, Optional, ParserPython, RegExMatch, ZeroOrMore
from jaxlib.pytree import PyTreeDef

__all__: List[str] = []


# Grammar
def g_tree() -> Any:
    return 'PyTreeDef(', g_type, ', ', g_values, ')'
def g_type() -> Any:
    return [g_dict, g_tuple, g_none, g_class]
def g_none() -> Any:
    return 'None'
def g_tuple() -> Any:
    return 'tuple'
def g_class() -> Any:
    return ("<class ", g_key, ">", g_crap)
def g_dict() -> Any:
    return ('dict[[', g_key, ZeroOrMore(',', g_key), ']]')
def g_crap() -> Any:
    return "[", RegExMatch(r"[^\[\]]*"), ZeroOrMore(g_crap), RegExMatch(r"[^\[\]]*"), "]"
def g_key() -> Any:
    return [RegExMatch("'[^']*'"),
            ('(', g_key, OneOrMore(', ', g_key), ')')]
def g_values() -> Any:
    return '[', Optional(g_value), ZeroOrMore(',', g_value), ']'
def g_value() -> Any:
    return [g_star, g_tree]
def g_star() -> Any:
    return '*'


indentation = 4


def display(r: str, result: Any, indent: int) -> str:
    def get(node: Any) -> str:
        return r[node.position: node.position_end]
    def unquote(s: str) -> str:
        quotes = "\"'"
        if s[0] not in quotes or s[-1] not in quotes:
            return s
        return s[1: -1]
    t = result[1][0]
    v = result[3]
    rn = t.rule_name[2:]
    keys = None
    if rn == 'class':
        class_qualname = unquote(get(t[1]))
        index = class_qualname.rfind('.')
        class_path = class_qualname[:index]
        class_name = class_qualname[index + 1:]
        module = importlib.import_module(class_path)
        this_class = getattr(module, class_name)
        if hasattr(this_class, 'tree_fields'):
            keys = this_class.tree_fields
        value_type = cf.blue(f"{class_name}")
    elif rn == 'dict':
        keys = [unquote(get(x)) for x in t.g_key]
        value_type = cf.yellow("dict")
    elif rn == 'tuple':
        value_type = cf.cyan("tuple")
    elif rn == 'none':
        value_type = cf.base1("None")
    else:
        assert False
    retval = value_type + "\n"
    if len(v) > 2:
        for i, x in enumerate(v.g_value):
            if x[0].rule_name == 'g_star':
                # retval += " " * (indent + indentation) + cf.base0('*') + "\n"
                continue
            retval += " " * (indent + indentation)
            if keys is not None:
                retval += cf.magenta(keys[i]) + cf.base00("=")
            retval += display(r, x[0], indent + indentation)
    return str(retval)


old_repr = PyTreeDef.__repr__


def new_repr(self: PyTreeDef) -> str:
    r = old_repr(self)
    parser = ParserPython(g_tree)
    try:
        result = parser.parse(r)
        return "\n" + display(r, result, 0)
    except:
        return r


PyTreeDef.__repr__ = new_repr
